#!/usr/bin/env python

# GNU GENERAL PUBLIC LICENSE (GPLv3)
# Copyright (c) 2020 - 2021 Bernardo Bulgarelli Labronici
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>5.


# JPEG Ghost in GIMP as described by
#  Hany Farid. Exposing digital forgeries from jpeg ghosts.
#  IEEE transactions on information forensics and security

# Feedback welcome: bernardobl@gmail.com

import os
import sys
#Current folder is not with sys path. cannot load GFUtils
sys.path.append(os.path.dirname(__file__))

from gimpfu import *
from array import array
from GFutils import image_info


def JPEG_GHOST(filepath,quality,norm,info,batch,step):
    #open image
    opImage = pdb.file_jpeg_load(filepath,filepath)

    if batch:
        qual = quality
        while qual < 1.00:
            DO_JPEG_GHOST(opImage,filepath,qual,norm,info)
            qual += step
    else:
        DO_JPEG_GHOST(opImage,filepath,quality,norm,info)

def DO_JPEG_GHOST(opImage,filepath,quality,norm,info):
    
    #open image
    layer = opImage.layers[0]
    width = layer.width
    height = layer.height


   #save lower quality
    temp = "GHOSTtemp.jpg"
    pdb.file_jpeg_save(opImage,layer,temp,temp,quality,0,0,0,'none',0,0,0,0)
   
    #Open lower quality image
    lower_img = pdb.file_jpeg_load(temp,temp)
    layer_q = lower_img.layers[0]

    #Get pixel from original image
    region_o = layer.get_pixel_rgn(0,0,width,height,False,False)
    #copy array for performance (up to 10x faster)
    pixels_o = array("B", region_o[0:width,0:height])    

    #Get pixels from lower quality
    region_q = layer_q.get_pixel_rgn(0,0,width,height,False,False)
    #copy array for performance (up to 10x faster)
    pixels_q = array("B", region_q[0:width,0:height])      

    
    #----------------------------------------
    #-----Create JPEG Ghost image image------
    #----------------------------------------
    
    #Image RGB with opacity channel (4 channels)
    alpha = 1
    type = RGB
    channels = 3
    
    pixels_num = region_o.bpp
    
    #check the real number of channels
    if pixels_num == 3: #RGB withou opacity
        alpha = 0
    elif pixels_num == 2: #Gray scale with opacity
       alpha = 1
       type = GRAY
       channels = 1
    elif pixels_num == 1: #gray scale
       alpha = 0
       type = GRAY
       channels = 1
        
    
    #create new image with the right number of channels    
    img = pdb.gimp_image_new(width, height, type)
    newLayer = img.new_layer("",pos=1,fill_mode=WHITE_FILL,alpha=alpha)
    
    region_n=newLayer.get_pixel_rgn(0, 0, width, height,True,False)
    
    #aray to hold all RGB (or gray scale) pixels values
    dest_pixels = array("B", "\x00" * (width * height * pixels_num))
    
    #Hold square error for pixels vlues (including RGB)
    squareError = [0.0] * width * height
 
    #Calculate square error once for each pixel
    for x in range( width * height):
        pos = x * pixels_num
        for y in range(channels):   
            squareError[x] += (pixels_o[pos  + y] - pixels_q[pos + y])**2
        
    #gamma array hold raw result of ghost for each pixel
    gamma = [0.0] * width * height
    minGamma = 0
    maxGamma = 0
    delta =  (channels * (norm**2))       
        
    for x in range( width * height):
        
        #Get the normalization windows around the pixel (return the index of pixels in box)
        index = getNormWindow(norm, width, height, x )     
        
        #------------------------------
        #------ Ghost Function --------
        #------------------------------
        
        for i in index:  
            gamma[x] += squareError[i]
        gamma[x] /=  delta##if norm greater than width or heigth?             
        
        #------------------------------
        
        
        #Get min and max gamma to normalize gamma (0 < gamma < 1.0)
        if x == 0:
          minGamma = gamma[x]
          maxGamma = gamma[x]
        else: 
          if gamma[x] < minGamma:
            minGamma = gamma[x]
          if gamma[x] > maxGamma:
            maxGamma = gamma[x]

    
    #Generate RGB image
    for x in range(width * height):
      
      #correct is the normalized gamma value (0 < gamma < 1.0)
      correct = (gamma[x] - minGamma)/(maxGamma - minGamma)
      
      pos = x*pixels_num
      
      #is GRAY Channel only.
      if channels == 1:
         dest_pixels[pos] = int(correct * 255)
      
      #Is RGB. Make the color gradient (R->G->B)
      else:  
         R = 0
         G = 0
         B = 0
         #from 1.0 to 0.5 grdient from Red to Green
         if correct > 0.5: 
           R = int(255*(correct - 0.5)/0.5)
           G = int(255 - 255*(correct - 0.5)/0.5)
         #from 0.5 to 0.0 gradient from Green to Blue  
         else:
            G = int(255*correct/0.5)
            B = int(255- 255*correct/0.5)
         
         dest_pixels[pos] = R
         dest_pixels[pos + 1] = G
         dest_pixels[pos + 2] = B

    #remove old image
    os.remove(temp)

    #past pixels in the new image and display it
    region_n[0:width,0:height]=dest_pixels.tostring() 


    if info: 
        image_info(img,"quality: " + str(quality*100) + "%\nnormalization: " + str(norm)) 
    
    
    display=pdb.gimp_display_new(img)
    
    
#Get the normalzed box around each pixel.
#Box has size "norm x norm"     
def getNormWindow(norm, width, height, pos):

   #convert from list of pixels to x,y coordinates
   x_pos = pos % width;
   y_pos = pos // width;
  
   #get the starting x coordinate of box
   if norm >= width:
       wXb = 0
       lX = width
   else:    
       wXb = (width-norm)*x_pos // (width - 1);
       lX = norm
   
   #get the end coordinate of box
   if norm >= height:
       wYb = 0
       lY = height
   else:
       wYb = (height-norm)*y_pos // (height - 1);
       lY = norm
   
   index = [0]*(norm**2)
   
   #create a list of references for the normalization box. this references is converted back to references in the pixels list
   i = 0
   for y in range(lY):
     for x in range(lX):
        index[i] = (wXb + x) + (wYb + y) * width
        i += 1  
         
   return index
    


register(
    "python-fu-jpeg-ghost",
    "Jpeg Ghost analysis",
    "Performe an JPEG Ghost analisys in the selected image. Abnormal compression level is highlighted",
    "Bernardo Bulgarelli Labronici", "Bernardo Bulgarelli Labronici", "2020",
    "Jpeg Ghost analysis",
    "", 
    [
        (PF_FILE, "filepath", "File to analyse", "/Default/Path"),
        (PF_FLOAT, "quality", "JPEG quality", 0.9),
        (PF_INT, "norm", "Normalize area in pixels", 16),
        (PF_BOOL, "info","Attach info to image",True),
        (PF_BOOL, "batch","Batch Render",False),
        (PF_FLOAT, "step","Step for Batch Render", 0.05)
    ],
    [],
    JPEG_GHOST, menu="<Image>/Forensics")

main()